# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2025

import torch
import copy
from torch import Tensor
import triton
import triton.language as tl
from ..dtypes import is_mx_dtype
from .config import AUTOTUNE
from .utils import *

KEYS = [
    "M_CLOSEST",
    "N",
    "K",
    "group_size",
    "elements_per_sample",
    "type_id",
    "a_sizeof",
    "b_sizeof",
]
MATMUL_TYPE = "GEMM"


def kernel_config_pruner(configs, nargs, **kwargs):
    from ..core import GEMLITE_TRITON_CONFIG_CACHE

    m = nargs["M"]
    n = nargs["N"]
    k = nargs["K"]
    g = nargs["group_size"]
    e = nargs["elements_per_sample"]
    t = nargs["type_id"]
    a_sizeof = nargs["a_sizeof"]
    b_sizeof = nargs["b_sizeof"]

    # Check cache
    if MATMUL_TYPE in GEMLITE_TRITON_CONFIG_CACHE:
        signature = str(tuple([get_closest_m(m), n, k, g, e, t]))
        if signature in GEMLITE_TRITON_CONFIG_CACHE[MATMUL_TYPE]:
            config = copy.deepcopy(GEMLITE_TRITON_CONFIG_CACHE[MATMUL_TYPE][signature])
            num_stages = config.pop("num_stages")
            num_warps = config.pop("num_warps")
            num_ctas = config.pop("num_ctas")

            config.pop("num_buffers_warp_spec", None)
            config.pop("num_consumer_groups", None)
            config.pop("reg_dec_producer", None)
            config.pop("reg_inc_consumer", None)
            config["NUM_STAGES"] = num_stages

            yield triton.Config(config, num_stages=num_stages, num_warps=num_warps)
            return

    gpu_shared_memory = get_gpu_shared_memory()
    load_scales_as_block = kwargs["load_scales_as_block"]
    used = set()
    for config in configs:
        group_size_m = config.kwargs["GROUP_SIZE_M"]
        block_size_m = config.kwargs["BLOCK_SIZE_M"]
        block_size_n = min(n, config.kwargs["BLOCK_SIZE_N"])
        block_size_k = min(k, config.kwargs["BLOCK_SIZE_K"])

        A_load_order = config.kwargs["A_load_order"]
        num_stages = config.num_stages
        num_warps = config.num_warps

        # Autotune prune the batch_size
        if m <= 16:
            block_size_m = 16
        elif m <= 32:
            block_size_m = min(max(block_size_m, 16), 32)  # m: [16...32]
        elif m <= 64:
            block_size_m = min(max(block_size_m, 32), 64)  # m: [32...64]
        elif m <= 128:
            block_size_m = min(max(block_size_m, 64), 128)  # m: [64...128]
        elif m <= 256:
            block_size_m = min(max(block_size_m, 64), 256)  # m: [128...256]
        elif m > 256:
            block_size_m = min(max(block_size_m, 64), 256)  # m > 256

        # Constraint: BLOCK_SIZE_K >= group_size, only for load_as_block = False
        if load_scales_as_block:
            num_stages = max(
                num_stages, 2
            )  # for dot_scaled kernels with pipelined loads
            if e > 1:
                block_size_k = max(block_size_k, 64)  # m16n8k64
            else:
                block_size_k = max(block_size_k, 32)  # m16n8k32
        else:
            block_size_k = min(block_size_k, g)

        block_size_k = next_power_of_2(block_size_k)
        block_size_n = next_power_of_2(block_size_n)

        # Hint: skip block_size_n > block_size_k for col-major non-packed data.

        # Nvidia
        if not IS_HIP:
            if e > 1 and not load_scales_as_block:
                # Limit num stages when data is packed
                num_stages = min(num_stages, 4)
            if e == 1 and num_stages == 1:
                # skip num_stages=1 for non-packed weights
                continue

        # Avoid OOM
        while num_stages > 0 and not load_scales_as_block:  # TODO: revisit MXFP case
            shared_mem = (
                block_size_m * block_size_k * a_sizeof
                + block_size_k * block_size_n * b_sizeof
            )
            if e > 1:
                shared_mem += block_size_k * block_size_n * a_sizeof
            shared_mem *= num_stages
            if int(shared_mem) <= gpu_shared_memory:
                break
            num_stages -= 1

        if num_stages == 0:
            continue  # config too large

        ###########################################
        if load_scales_as_block:  # tmp MXFP fix
            block_size_k = min(block_size_k, 256)
        ###########################################

        key = (
            block_size_m,
            block_size_n,
            block_size_k,
            group_size_m,
            A_load_order,
            num_stages,
            num_warps,
        )

        new_config = {
            "BLOCK_SIZE_M": block_size_m,
            "BLOCK_SIZE_N": block_size_n,
            "BLOCK_SIZE_K": block_size_k,
            "GROUP_SIZE_M": group_size_m,
            "A_load_order": A_load_order,
            "NUM_STAGES": num_stages,
        }

        if IS_HIP:
            new_config["waves_per_eu"] = config.kwargs.get("waves_per_eu", 0)
            new_config["matrix_instr_nonkdim"] = config.kwargs.get(
                "matrix_instr_nonkdim", 16
            )  # MI300X
            key = key + (new_config["waves_per_eu"], new_config["matrix_instr_nonkdim"])

        if key in used:
            continue

        used.add(key)
        yield triton.Config(new_config, num_stages=num_stages, num_warps=num_warps)


########################################################################################################################################################################
# Nvidia
def get_max_autotune_config_nvidia():
    stages = [1, 4, 5] if gpu_has_more_shared_memory() else [1, 2, 4]
    configs = []
    for A in [0, 2]:
        for w in [4, 8]:
            for s in stages:
                for M in [16, 32, 64, 128, 256]:
                    for N in [32, 64, 128, 256, 512]:
                        for K in [32, 64, 128, 256, 512]:
                            configs.append(
                                triton.Config(
                                    {
                                        "BLOCK_SIZE_M": M,
                                        "BLOCK_SIZE_N": N,
                                        "BLOCK_SIZE_K": K,
                                        "GROUP_SIZE_M": 8,
                                        "A_load_order": A,
                                    },
                                    num_warps=w,
                                    num_stages=s,
                                )
                            )

    return configs


def get_fast_autotune_config_nvidia():
    configs = []  # BLOCK_SIZE_M is automatically adapted in the config pruning.
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
                "A_load_order": 2,
            },
            num_warps=4,
            num_stages=4,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 64,
                "GROUP_SIZE_M": 8,
                "A_load_order": 2,
            },
            num_warps=4,
            num_stages=4,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "A_load_order": 2,
            },
            num_warps=8,
            num_stages=4,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 256,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
            },
            num_warps=4,
            num_stages=5,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
            },
            num_warps=4,
            num_stages=4,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 64,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
            },
            num_warps=4,
            num_stages=4,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
            },
            num_warps=8,
            num_stages=5,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 256,
                "GROUP_SIZE_M": 8,
                "A_load_order": 2,
            },
            num_warps=8,
            num_stages=4,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 128,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
                "A_load_order": 2,
            },
            num_warps=8,
            num_stages=5,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 128,
                "BLOCK_SIZE_K": 64,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
            },
            num_warps=4,
            num_stages=5,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 128,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "A_load_order": 2,
            },
            num_warps=4,
            num_stages=4,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 128,
                "BLOCK_SIZE_K": 256,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
            },
            num_warps=4,
            num_stages=4,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 256,
                "BLOCK_SIZE_K": 64,
                "GROUP_SIZE_M": 8,
                "A_load_order": 2,
            },
            num_warps=8,
            num_stages=4,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 256,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
            },
            num_warps=8,
            num_stages=4,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 512,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "A_load_order": 2,
            },
            num_warps=8,
            num_stages=3,
        )
    )
    return configs


def get_default_config_nvidia():
    return [
        triton.Config(
            {
                "BLOCK_SIZE_M": 16,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "NUM_STAGES": 4,
            },
            num_warps=4,
            num_stages=4,
        ),
    ]


########################################################################################################################################################################
# AMD - Instinct MI300X


def get_max_autotune_config_amd():
    configs = []
    for A in [0]:
        for w in [4, 8]:
            for s in [1, 2]:
                for v in [0, 2, 4]:
                    for M in [16, 32, 64, 128, 256]:
                        for N in [32, 64, 128, 256, 512]:
                            for K in [32, 64, 128, 256, 512]:
                                configs.append(
                                    triton.Config(
                                        {
                                            "BLOCK_SIZE_M": M,
                                            "BLOCK_SIZE_N": N,
                                            "BLOCK_SIZE_K": K,
                                            "GROUP_SIZE_M": 8,
                                            "A_load_order": A,
                                            "waves_per_eu": v,
                                        },
                                        num_warps=w,
                                        num_stages=s,
                                    )
                                )

    return configs


def get_fast_autotune_config_amd():
    configs = []  # BLOCK_SIZE_M is automatically adapted in the config pruning.
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 64,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 4,
            },
            num_warps=4,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 64,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 0,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 2,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 256,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 2,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 512,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 0,
            },
            num_warps=8,
            num_stages=2,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 2,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 64,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 0,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 64,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=1,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 4,
            },
            num_warps=4,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 0,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 256,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 4,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 256,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 0,
            },
            num_warps=8,
            num_stages=2,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 128,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 0,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 128,
                "BLOCK_SIZE_K": 64,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 128,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 0,
            },
            num_warps=8,
            num_stages=2,
        )
    )
    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 128,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=2,
        )
    )

    configs.append(
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 256,
                "BLOCK_SIZE_K": 128,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "waves_per_eu": 2,
            },
            num_warps=4,
            num_stages=1,
        )
    )
    return configs


def get_default_config_amd():
    return [
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
                "A_load_order": 0,
                "NUM_STAGES": 2,
            },
            num_warps=4,
            num_stages=2,
        ),
    ]


########################################################################################################################################################################
if IS_HIP:
    get_max_autotune_config = get_max_autotune_config_amd
    get_fast_autotune_config = get_fast_autotune_config_amd
    get_default_config = get_default_config_amd
else:
    get_max_autotune_config = get_max_autotune_config_nvidia
    get_fast_autotune_config = get_fast_autotune_config_nvidia
    get_default_config = get_default_config_nvidia

AUTOTUNE_SETTING = AUTOTUNE.GEMM
if AUTOTUNE_SETTING == "max":
    get_autotune_config = get_max_autotune_config
elif AUTOTUNE_SETTING == "fast":
    get_autotune_config = get_fast_autotune_config
else:
    get_autotune_config = get_default_config


@triton.autotune(
    configs=get_autotune_config(),
    key=KEYS,
    prune_configs_by={"early_config_prune": kernel_config_pruner},
    use_cuda_graph=AUTOTUNE.USE_CUDA_GRAPH,
)
@triton.jit
def gemm_INT_kernel(
    a_ptr,
    b_ptr,
    c_ptr,
    scales_ptr,
    zeros_ptr,
    scales_a_ptr,
    M,
    N,
    K,
    M_CLOSEST,
    ######### Quant parms #########
    W_nbits: tl.constexpr,
    group_size: tl.constexpr,
    unpack_mask: tl.constexpr,
    elements_per_sample: tl.constexpr,
    #################################
    type_id: tl.constexpr,
    a_sizeof: tl.constexpr,
    b_sizeof: tl.constexpr,
    ######### Strides #########
    stride_am,
    stride_ak,
    stride_bk,
    stride_bn,
    stride_cm,
    stride_cn,
    stride_meta_a_m,
    stride_meta_a_g,
    stride_meta_g,
    stride_meta_n,
    ######### Dtypes #########
    load_scales_as_block,  # False
    input_dtype: tl.constexpr,
    output_dtype: tl.constexpr,
    acc_dtype: tl.constexpr,
    meta_dtype: tl.constexpr,
    ######### Meta-data mode #########
    channel_scale_mode: tl.constexpr,
    W_group_mode: tl.constexpr,
    zero_is_scalar: tl.constexpr,
    ######### tuning params #########
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    BLOCK_SIZE_K: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
    NUM_STAGES: tl.constexpr,
    A_load_order: tl.constexpr,
    data_contiguous: tl.constexpr,
    #################################
    meta_evict_policy: tl.constexpr = "",
    a_evict: tl.constexpr = "",
    b_evict: tl.constexpr = "",
):
    """
    Based on https://github.com/fpgaminer/GPTQ-triton
    GEMM for C = matmul(A, dequantize(B, scales, zeros))
    A is of shape (M, K): float16 or bfloat16
    B is of shape (K//elements_per_sample, N): int32 as a packed matrix
    C is of shape (M, N): float16 or bfloat16 depending on the input A
    scales and zeros is of shape (group_size, N): float16 or bfloat16

    BLOCK_SIZE_M >=16
    BLOCK_SIZE_K <= group_size
    """

    pid = tl.program_id(axis=0)

    # Swizzle?
    if elements_per_sample > 1:
        pid_m, pid_n = linear_tile(pid, M, N, BLOCK_SIZE_M, BLOCK_SIZE_N, None)
    else:
        pid_m, pid_n = swizzle_tile(pid, M, N, BLOCK_SIZE_M, BLOCK_SIZE_N, GROUP_SIZE_M)

    num_pid_m = tl.cdiv(M, BLOCK_SIZE_M)
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)
    num_pid_k = tl.cdiv(K, BLOCK_SIZE_K)

    # Offsets
    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    offs_k = tl.arange(0, BLOCK_SIZE_K)

    # Offsets
    #############################################################################################################
    if data_contiguous:
        offs_bn = offs_n
    else:
        offs_bn = tl.max_contiguous(tl.multiple_of(offs_n, BLOCK_SIZE_N), BLOCK_SIZE_N)
    offs_am = tl.max_contiguous(tl.multiple_of(offs_m, BLOCK_SIZE_M), BLOCK_SIZE_M)
    offs_ak = offs_k
    offs_bk = offs_k

    b_ptrs = b_ptr + (
        (offs_bk[:, None] // elements_per_sample) * stride_bk
        + offs_bn[None, :] * stride_bn
    )
    q_shift = ((offs_bk % elements_per_sample) * W_nbits).to(tl.int32)[:, None]

    # Inputs
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_ak[None, :] * stride_ak)
    a_mask = ((offs_am[:, None] < M) & (offs_ak[None, :] < K)).to(tl.int1)

    # Meta data stuff
    scales_ptrs = scales_ptr + offs_bn[None, :] * stride_meta_n
    zeros_ptrs = zeros_ptr + offs_bn[None, :] * stride_meta_n

    stride_mul: tl.constexpr = BLOCK_SIZE_K / group_size
    BLOCK_SIZE_K_P: tl.constexpr = BLOCK_SIZE_K // elements_per_sample

    if zero_is_scalar:
        zero_scalar = tl.load(zeros_ptr, eviction_policy="evict_last")

    #############################################################################################################
    acc = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=acc_dtype)

    for k in range(num_pid_k):
        if A_load_order == 0:  # Early load
            a = tl.load(a_ptrs, mask=a_mask, other=0.0, eviction_policy=a_evict)

        b = tl.load(b_ptrs, eviction_policy=b_evict)

        if A_load_order == 1:  # Early load
            a = tl.load(a_ptrs, mask=a_mask, other=0.0, eviction_policy=a_evict)

        # Meta-data loading policy
        if W_group_mode > 0:
            k_m = (k * stride_mul).to(tl.int32)

        if W_group_mode >= 2:  # [2, 3, 4]
            scales = tl.load(
                scales_ptrs + k_m * stride_meta_g, eviction_policy=meta_evict_policy
            )
        else:
            scales = None

        if W_group_mode == 1 or W_group_mode >= 3:  # [1, 3, 4]
            if zero_is_scalar:
                zeros = zero_scalar
            else:
                zeros = tl.load(
                    zeros_ptrs + k_m * stride_meta_g, eviction_policy=meta_evict_policy
                )
        else:
            zeros = None

        if A_load_order == 2:  # Mid load
            a = tl.load(a_ptrs, mask=a_mask, other=0.0, eviction_policy=a_evict)

        # Unpack and dequantize
        b = dequantize(
            b,
            scales,
            zeros,
            q_shift,
            meta_dtype,
            unpack_mask,
            elements_per_sample,
            W_group_mode,
            zero_is_scalar,
        )

        if A_load_order == 3:  # Late load
            a = tl.load(a_ptrs, mask=a_mask, other=0.0, eviction_policy=a_evict)

        # Dot
        acc = tl.dot(a, b.to(input_dtype), acc=acc, out_dtype=acc_dtype)

        # Advance
        a_ptrs += BLOCK_SIZE_K * stride_ak
        b_ptrs += BLOCK_SIZE_K_P * stride_bk

    #############################################################################################################
    # Channel-wise scaling
    if channel_scale_mode == 1:  # weight-only
        scales_b = tl.load(
            scales_ptr + offs_bn,
            mask=offs_bn < N,
            other=1,
            eviction_policy=meta_evict_policy,
        )
        acc = acc.to(meta_dtype) * scales_b[None, :]

    if channel_scale_mode == 2:  # activation-only
        scales_a = tl.load(
            scales_a_ptr + offs_am,
            mask=offs_am < M,
            other=1,
            eviction_policy=meta_evict_policy,
        )
        scales_b = tl.full((BLOCK_SIZE_N,), value=1, dtype=meta_dtype)
        acc = acc.to(meta_dtype) * (scales_a[:, None] * scales_b[None, :])

    if channel_scale_mode == 3:  # weight + activation
        scales_a = tl.load(
            scales_a_ptr + offs_am,
            mask=offs_am < M,
            other=1,
            eviction_policy=meta_evict_policy,
        )
        scales_b = tl.load(
            scales_ptr + offs_bn,
            mask=offs_bn < N,
            other=1,
            eviction_policy=meta_evict_policy,
        )
        acc = acc.to(meta_dtype) * (scales_a[:, None] * scales_b[None, :])

    acc = acc.to(output_dtype)
    #############################################################################################################
    # Output
    offs_cm = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    offs_cn = tl.max_contiguous(tl.multiple_of(offs_cn, BLOCK_SIZE_N), BLOCK_SIZE_N)
    c_ptrs = c_ptr + (offs_cm[:, None] * stride_cm + offs_cn[None, :] * stride_cn)
    tl.store(c_ptrs, acc, mask=(offs_cm[:, None] < M) & (offs_cn[None, :] < N))


@triton.autotune(
    configs=get_autotune_config(),
    key=KEYS,
    prune_configs_by={"early_config_prune": kernel_config_pruner},
    use_cuda_graph=AUTOTUNE.USE_CUDA_GRAPH,
)
@triton.jit
def gemm_MX_kernel(
    a_ptr,
    b_ptr,
    c_ptr,
    scales_ptr,
    zeros_ptr,
    scales_a_ptr,
    M,
    N,
    K,
    M_CLOSEST,
    ######### Quant parms #########
    W_nbits: tl.constexpr,
    group_size: tl.constexpr,
    unpack_mask: tl.constexpr,
    elements_per_sample: tl.constexpr,
    #################################
    type_id: tl.constexpr,
    a_sizeof: tl.constexpr,
    b_sizeof: tl.constexpr,
    ######### Strides #########
    stride_am,
    stride_ak,
    stride_bk,
    stride_bn,
    stride_cm,
    stride_cn,
    stride_meta_a_m: tl.constexpr,
    stride_meta_a_g: tl.constexpr,
    stride_meta_n: tl.constexpr,
    stride_meta_g: tl.constexpr,
    ######### Dtypes #########
    load_scales_as_block,  # True
    input_dtype: tl.constexpr,
    output_dtype: tl.constexpr,
    meta_dtype: tl.constexpr,
    acc_dtype: tl.constexpr,
    ######### Meta-data mode #########
    channel_scale_mode: tl.constexpr,
    W_group_mode: tl.constexpr,
    zero_is_scalar: tl.constexpr,
    ######### tuning params #########
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    BLOCK_SIZE_K: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
    NUM_STAGES: tl.constexpr,
    A_load_order: tl.constexpr,
    data_contiguous: tl.constexpr,
    #################################
    meta_evict_policy: tl.constexpr = "",
    a_evict: tl.constexpr = "",
    b_evict: tl.constexpr = "",
    meta_scale_norm: tl.constexpr = (0.05**2),
):

    pid = tl.program_id(axis=0)
    pid_m, pid_n = swizzle_tile(pid, M, N, BLOCK_SIZE_M, BLOCK_SIZE_N, GROUP_SIZE_M)
    num_pid_k = tl.cdiv(K, BLOCK_SIZE_K)

    a_ptr_dtype: tl.constexpr = a_ptr.dtype.element_ty
    if a_ptr_dtype == tl.float16:
        a_dtype: tl.constexpr = "fp16"
        elements_per_sample_a: tl.constexpr = 1
    if a_ptr_dtype == tl.bfloat16:
        a_dtype: tl.constexpr = "bf16"
        elements_per_sample_a: tl.constexpr = 1
    if a_ptr_dtype == tl.float8e4nv:
        a_dtype: tl.constexpr = "e4m3"
        elements_per_sample_a: tl.constexpr = 1
    if a_ptr_dtype == tl.uint8:
        a_dtype: tl.constexpr = "e2m1"  # FP4
        elements_per_sample_a: tl.constexpr = 2

    if elements_per_sample == 1:  # FP8
        b_dtype: tl.constexpr = "e4m3"
    if elements_per_sample == 2:  # FP4
        b_dtype: tl.constexpr = "e2m1"

    # A
    BLOCK_SIZE_K_A: tl.constexpr = BLOCK_SIZE_K // elements_per_sample_a
    offs_am = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_ak = tl.arange(0, BLOCK_SIZE_K_A)
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_ak[None, :] * stride_ak)
    a_mask = (
        (offs_am[:, None] < M) & (offs_ak[None, :] < K // elements_per_sample_a)
    ).to(tl.int1)

    # B
    BLOCK_SIZE_K_B: tl.constexpr = BLOCK_SIZE_K // elements_per_sample
    offs_bk = tl.arange(0, BLOCK_SIZE_K_B)
    offs_bn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    b_ptrs = b_ptr + offs_bk[:, None] * stride_bk + offs_bn[None, :] * stride_bn

    # Scales
    stride_mul: tl.constexpr = BLOCK_SIZE_K / group_size
    BLOCK_SIZE_K_S: tl.constexpr = BLOCK_SIZE_K // group_size
    offs_k_scales = tl.arange(0, BLOCK_SIZE_K_S)
    offs_n_b_scales = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    scales_b_ptrs = (
        scales_ptr
        + offs_n_b_scales[:, None] * stride_meta_n
        + offs_k_scales[None, :] * stride_meta_g
    )  # [BLOCK_SIZE_N, BLOCK_SIZE_K // group_size]

    # B-scales
    if channel_scale_mode == 4:
        scales_a_ptrs = (
            scales_a_ptr
            + offs_am[:, None] * stride_meta_a_m
            + offs_k_scales[None, :] * stride_meta_a_g
        )

    acc = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=acc_dtype)
    for k in tl.range(num_pid_k, num_stages=NUM_STAGES):
        a = tl.load(a_ptrs, mask=a_mask, other=0.0, eviction_policy=a_evict)
        b = tl.load(b_ptrs, eviction_policy=b_evict)

        k_m = k * BLOCK_SIZE_K_S
        scales_b = tl.load(
            scales_b_ptrs + k_m * stride_meta_g, eviction_policy=meta_evict_policy
        )

        if channel_scale_mode == 4:
            scales_a = tl.load(
                scales_a_ptrs + k_m * stride_meta_a_g, eviction_policy=meta_evict_policy
            )
        else:
            scales_a = tl.full(
                (BLOCK_SIZE_M, BLOCK_SIZE_K_S), value=127, dtype=tl.uint8
            )

        acc = tl.dot_scaled(a, scales_a, a_dtype, b, scales_b, b_dtype, acc)

        a_ptrs += BLOCK_SIZE_K_A * stride_ak
        b_ptrs += BLOCK_SIZE_K_B * stride_bk

    # NVFP4 meta-scale
    if group_size == 16:
        acc *= meta_scale_norm

    #############################################################################################################
    # Channel-wise scaling
    if channel_scale_mode == 2:  # activation-only
        dtype: tl.constexpr = c_ptr.dtype.element_ty
        scales_a = tl.load(
            scales_a_ptr + offs_am,
            mask=offs_am < M,
            other=1,
            eviction_policy=meta_evict_policy,
        )
        scales_b = tl.full((BLOCK_SIZE_N,), value=1, dtype=dtype)
        acc = acc.to(dtype) * (scales_a[:, None] * scales_b[None, :])

    #############################################################################################################
    # Output
    offs_cm = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    c_ptrs = c_ptr + (offs_cm[:, None] * stride_cm + offs_cn[None, :] * stride_cn)
    mask = ((offs_cm[:, None] < M) & (offs_cn[None, :] < N)).to(tl.int1)
    tl.store(c_ptrs, acc, mask=mask)


def gemm_forward(
    x: Tensor,
    W_q: Tensor,
    scales: Tensor,
    zeros: Tensor,
    scales_x: Tensor,
    W_nbits: int,
    group_size: int,
    unpack_mask: int,
    elements_per_sample: int,
    input_dtype: int,
    output_dtype: int,
    acc_dtype: int,
    meta_dtype: int,
    channel_scale_mode: int,
    W_group_mode: int,
    data_contiguous: bool,
    type_id: int,
) -> Tensor:

    M, K, N = x.shape[0], W_q.shape[0] * elements_per_sample, W_q.shape[1]
    M_CLOSEST = get_closest_m(M)

    # assert K == W_q.shape[0] * elements_per_sample, "Invalid Input Shapes"
    output = torch.empty((M, N), device=W_q.device, dtype=DTYPE_TO_TORCH[output_dtype])

    grid = lambda META: (
        triton.cdiv(M, META["BLOCK_SIZE_M"]) * triton.cdiv(N, META["BLOCK_SIZE_N"]),
    )

    if scales_x is not None:
        stride_meta_a_m, stride_meta_a_g = scales_x.stride(0), scales_x.stride(1)
    else:
        stride_meta_a_m, stride_meta_a_g = None, None

    if is_mx_dtype(input_dtype):
        gemm_kernel = gemm_MX_kernel
        load_scales_as_block = True
    else:
        gemm_kernel = gemm_INT_kernel
        load_scales_as_block = False

    gemm_kernel[grid](
        x,
        W_q,
        output,
        scales,
        zeros,
        scales_x,
        M,
        N,
        K,
        M_CLOSEST,
        #############################################
        W_nbits,
        group_size,
        unpack_mask,
        elements_per_sample,
        type_id,
        x.dtype.itemsize,
        W_q.dtype.itemsize,
        ###############################################
        x.stride(0),
        x.stride(1),
        W_q.stride(0),
        W_q.stride(1),
        output.stride(0),
        output.stride(1),
        stride_meta_a_m,
        stride_meta_a_g,
        scales.stride(0),
        scales.stride(1),
        ################################################
        load_scales_as_block=load_scales_as_block,
        input_dtype=DTYPE_TO_TRITON[input_dtype],
        output_dtype=TORCH_DTYPE_TO_TRITON[output.dtype],
        acc_dtype=DTYPE_TO_TRITON[acc_dtype],
        meta_dtype=DTYPE_TO_TRITON[meta_dtype],
        ################################################
        channel_scale_mode=channel_scale_mode,
        W_group_mode=W_group_mode,
        zero_is_scalar=zeros.numel() == 1,
        data_contiguous=data_contiguous,
    )

    return output


class gemm:
    kernel = [gemm_INT_kernel, gemm_MX_kernel]
    forward = gemm_forward
    matmul_type = MATMUL_TYPE


__all__ = ["gemm"]



# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2025

import sys

if sys.version_info < (3, 12):
    import imp
else:
    import importlib as imp
from typing import Union

MATMUL_DTYPES = ["GEMV", "GEMV_REVSPLITK", "GEMV_SPLITK", "GEMM_SPLITK", "GEMM"]


class AUTOTUNE:
    GEMV = "fast"  # "max", "fast", "default"
    GEMV_REVSPLITK = "fast"
    GEMV_SPLITK = "fast"
    GEMM_SPLITK = "fast"
    GEMM = "fast"
    USE_CUDA_GRAPH = False


class KERNEL:
    ENABLE_CACHING = False
    CACHE_SIZE = 256


def reload_all_modules():
    # Avoid circular imports
    from . import gemm_kernels
    from . import gemm_splitK_kernels
    from . import gemm_splitK_persistent_kernels
    from . import gemv_kernels
    from . import gemv_splitK_kernels
    from . import gemv_revsplitK_kernels

    imp.reload(gemm_kernels)
    imp.reload(gemm_splitK_kernels)
    imp.reload(gemm_splitK_persistent_kernels)
    imp.reload(gemv_kernels)
    imp.reload(gemv_splitK_kernels)
    imp.reload(gemv_revsplitK_kernels)


def set_kernel_caching(enable: bool):
    KERNEL.ENABLE_CACHING = enable


def set_autotune(config: Union[dict, str, bool], **kwargs):
    if type(config) == str:
        for key in MATMUL_DTYPES:
            setattr(AUTOTUNE, key, config.lower())

    if type(config) == bool:
        for key in MATMUL_DTYPES:
            setattr(AUTOTUNE, key, "max" if config else "default")

    if type(config) == dict:
        for key in config:
            setattr(AUTOTUNE, key, config[key])

    if "use_cuda_graph" in kwargs:
        setattr(AUTOTUNE, "USE_CUDA_GRAPH", bool(kwargs["use_cuda_graph"]))

    reload_all_modules()



from transformers import Wav2Vec2Config, Wav2Vec2Model
from transformers.modeling_outputs import BaseModelOutput
import torch
import torch.nn.functional as F


def get_mask_from_lengths(lengths, max_len=None):
    lengths = lengths.to(torch.long)
    if max_len is None:
        max_len = torch.max(lengths).item()

    ids = (
        torch.arange(0, max_len)
        .unsqueeze(0)
        .expand(lengths.shape[0], -1)
        .to(lengths.device)
    )
    mask = ids < lengths.unsqueeze(1).expand(-1, max_len)

    return mask


def linear_interpolation(features, seq_len):
    features = features.transpose(1, 2)
    output_features = F.interpolate(
        features, size=seq_len, align_corners=True, mode="linear"
    )
    return output_features.transpose(1, 2)


# the implementation of Wav2Vec2Model is borrowed from
# https://github.com/huggingface/transformers/blob/HEAD/src/transformers/models/wav2vec2/modeling_wav2vec2.py
# initialize our encoder with the pre-trained wav2vec 2.0 weights.
class Wav2Vec2Model(Wav2Vec2Model):
    def __init__(self, config: Wav2Vec2Config):
        super().__init__(config)

    def forward(
        self,
        input_values,
        seq_len,
        attention_mask=None,
        mask_time_indices=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        self.config.output_attentions = False

        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        extract_features = self.feature_extractor(input_values)
        extract_features = extract_features.transpose(1, 2)
        extract_features = linear_interpolation(extract_features, seq_len=seq_len)

        if attention_mask is not None:
            # compute reduced attention_mask corresponding to feature vectors
            attention_mask = self._get_feature_vector_attention_mask(
                extract_features.shape[1], attention_mask, add_adapter=False
            )

        hidden_states, extract_features = self.feature_projection(extract_features)
        hidden_states = self._mask_hidden_states(
            hidden_states,
            mask_time_indices=mask_time_indices,
            attention_mask=attention_mask,
        )

        encoder_outputs = self.encoder(
            hidden_states,
            attention_mask=attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        hidden_states = encoder_outputs[0]

        if self.adapter is not None:
            hidden_states = self.adapter(hidden_states)

        if not return_dict:
            return (hidden_states,) + encoder_outputs[1:]
        return BaseModelOutput(
            last_hidden_state=hidden_states,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
        )

    def feature_extract(
        self,
        input_values,
        seq_len,
    ):
        extract_features = self.feature_extractor(input_values)
        extract_features = extract_features.transpose(1, 2)
        extract_features = linear_interpolation(extract_features, seq_len=seq_len)

        return extract_features

    def encode(
        self,
        extract_features,
        attention_mask=None,
        mask_time_indices=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        self.config.output_attentions = False

        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        if attention_mask is not None:
            # compute reduced attention_mask corresponding to feature vectors
            attention_mask = self._get_feature_vector_attention_mask(
                extract_features.shape[1], attention_mask, add_adapter=False
            )

        hidden_states, extract_features = self.feature_projection(extract_features)
        hidden_states = self._mask_hidden_states(
            hidden_states,
            mask_time_indices=mask_time_indices,
            attention_mask=attention_mask,
        )

        encoder_outputs = self.encoder(
            hidden_states,
            attention_mask=attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        hidden_states = encoder_outputs[0]

        if self.adapter is not None:
            hidden_states = self.adapter(hidden_states)

        if not return_dict:
            return (hidden_states,) + encoder_outputs[1:]
        return BaseModelOutput(
            last_hidden_state=hidden_states,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
        )



import copy
from contextlib import contextmanager

import torch
from loguru import logger


@contextmanager
def init_weights_on_device(device=torch.device("meta"), include_buffers: bool = False):
    old_register_parameter = torch.nn.Module.register_parameter
    if include_buffers:
        old_register_buffer = torch.nn.Module.register_buffer

    def register_empty_parameter(module, name, param):
        old_register_parameter(module, name, param)
        if param is not None:
            param_cls = type(module._parameters[name])
            kwargs = module._parameters[name].__dict__
            kwargs["requires_grad"] = param.requires_grad
            module._parameters[name] = param_cls(
                module._parameters[name].to(device), **kwargs
            )

    def register_empty_buffer(module, name, buffer, persistent=True):
        old_register_buffer(module, name, buffer, persistent=persistent)
        if buffer is not None:
            module._buffers[name] = module._buffers[name].to(device)

    def patch_tensor_constructor(fn):
        def wrapper(*args, **kwargs):
            kwargs["device"] = device
            return fn(*args, **kwargs)

        return wrapper

    if include_buffers:
        tensor_constructors_to_patch = {
            torch_function_name: getattr(torch, torch_function_name)
            for torch_function_name in ["empty", "zeros", "ones", "full"]
        }
    else:
        tensor_constructors_to_patch = {}

    try:
        torch.nn.Module.register_parameter = register_empty_parameter
        if include_buffers:
            torch.nn.Module.register_buffer = register_empty_buffer
        for torch_function_name in tensor_constructors_to_patch.keys():
            setattr(
                torch,
                torch_function_name,
                patch_tensor_constructor(getattr(torch, torch_function_name)),
            )
        yield
    finally:
        torch.nn.Module.register_parameter = old_register_parameter
        if include_buffers:
            torch.nn.Module.register_buffer = old_register_buffer
        for (
            torch_function_name,
            old_torch_function,
        ) in tensor_constructors_to_patch.items():
            setattr(torch, torch_function_name, old_torch_function)


def cast_to(
    weight: torch.Tensor, dtype: torch.dtype, device: torch.device
) -> torch.Tensor:
    if hasattr(weight, "__class__") and "torchao.dtypes" in str(weight.__class__):
        r = weight.to(device=device)
    elif hasattr(weight, "__class__") and "torchao.quantization" in str(
        weight.__class__
    ):
        r = weight.to(dtype=dtype, device=device)
    else:
        r = torch.empty_like(weight, dtype=dtype, device=device)
        r.copy_(weight)
    return r


def cast_to_device(weight: torch.Tensor, device: torch.device) -> torch.Tensor:
    if hasattr(weight, "__class__") and "optimum.quanto" in str(weight.__class__):
        r = weight.to(device)
        return r
    else:
        r = torch.empty_like(weight, device=device)
        r.copy_(weight)
        return r


class AutoWrappedModule(torch.nn.Module):
    def __init__(
        self,
        module: torch.nn.Module,
        offload_dtype,
        offload_device,
        onload_dtype,
        onload_device,
        computation_dtype,
        computation_device,
    ):
        super().__init__()
        self.module = module.to(dtype=offload_dtype, device=offload_device)
        self.offload_dtype = offload_dtype
        self.offload_device = offload_device
        self.onload_dtype = onload_dtype
        self.onload_device = onload_device
        self.computation_dtype = computation_dtype
        self.computation_device = computation_device
        self.state = 0

    def offload(self):
        if self.state == 1 and (
            self.offload_dtype != self.onload_dtype
            or self.offload_device != self.onload_device
        ):
            self.module.to(dtype=self.offload_dtype, device=self.offload_device)
            self.state = 0

    def onload(self):
        if self.state == 0 and (
            self.offload_dtype != self.onload_dtype
            or self.offload_device != self.onload_device
        ):
            self.module.to(dtype=self.onload_dtype, device=self.onload_device)
            self.state = 1

    def forward(self, *args, **kwargs):
        if (
            self.onload_dtype == self.computation_dtype
            and self.onload_device == self.computation_device
        ):
            module = self.module
        else:
            module = copy.deepcopy(self.module).to(
                dtype=self.computation_dtype, device=self.computation_device
            )
        r = module(*args, **kwargs)
        return r

    def __getattr__(self, name):
        # Proxy attribute access to wrapped module for non-special attributes
        if name in [
            "module",
            "offload_dtype",
            "offload_device",
            "onload_dtype",
            "onload_device",
            "computation_dtype",
            "computation_device",
            "state",
        ]:
            return super().__getattr__(name)
        return getattr(self.module, name)


class AutoWrappedLinear(torch.nn.Linear):
    def __init__(
        self,
        module: torch.nn.Linear,
        offload_dtype,
        offload_device,
        onload_dtype,
        onload_device,
        computation_dtype,
        computation_device,
    ):
        with init_weights_on_device(device=torch.device("meta")):
            super().__init__(
                in_features=module.in_features,
                out_features=module.out_features,
                bias=module.bias is not None,
                dtype=offload_dtype,
                device=offload_device,
            )
        self.weight = module.weight
        self.bias = module.bias
        self.offload_dtype = offload_dtype
        self.offload_device = offload_device
        self.onload_dtype = onload_dtype
        self.onload_device = onload_device
        self.computation_dtype = computation_dtype
        self.computation_device = computation_device
        self.state = 0

    def offload(self):
        if self.state == 1 and (
            self.offload_dtype != self.onload_dtype
            or self.offload_device != self.onload_device
        ):
            self.to(dtype=self.offload_dtype, device=self.offload_device)
            self.state = 0

    def onload(self):
        if self.state == 0 and (
            self.offload_dtype != self.onload_dtype
            or self.offload_device != self.onload_device
        ):
            self.to(dtype=self.onload_dtype, device=self.onload_device)
            self.state = 1

    def forward(self, x, *args, **kwargs):
        if (
            self.onload_dtype == self.computation_dtype
            and self.onload_device == self.computation_device
        ):
            weight, bias = self.weight, self.bias
        else:
            weight = cast_to(
                self.weight, self.computation_dtype, self.computation_device
            )
            bias = (
                None
                if self.bias is None
                else cast_to(self.bias, self.computation_dtype, self.computation_device)
            )
        r = torch.nn.functional.linear(x, weight, bias)
        return r


def enable_vram_management_recursively(
    model: torch.nn.Module,
    module_map: dict,
    module_config: dict,
    max_num_param=None,
    overflow_module_config: dict = None,
    total_num_param=0,
):
    for name, module in model.named_children():
        module_type = module.__class__.__name__
        for source_module, target_module in module_map.items():
            if isinstance(module, source_module):
                num_param = sum(p.numel() for p in module.parameters())
                is_overflow = (
                    max_num_param is not None
                    and total_num_param + num_param > max_num_param
                )
                if is_overflow:
                    module_config_ = overflow_module_config
                else:
                    module_config_ = module_config
                module_ = target_module(module, **module_config_)
                setattr(model, name, module_)
                total_num_param += num_param
                break
        else:
            total_num_param = enable_vram_management_recursively(
                module,
                module_map,
                module_config,
                max_num_param,
                overflow_module_config,
                total_num_param,
            )
    return total_num_param


def enable_vram_management(
    model: torch.nn.Module,
    module_map: dict,
    module_config: dict,
    max_num_param=None,
    overflow_module_config: dict = None,
):
    total_num_param = enable_vram_management_recursively(
        model,
        module_map,
        module_config,
        max_num_param,
        overflow_module_config,
        total_num_param=0,
    )
    model.vram_management_enabled = True
    max_persistent_str = f"{max_num_param / 1e6:.2f}M" if max_num_param else "None"
    logger.info(
        f"[VRAM] Total wrapped parameters: {total_num_param / 1e6:.2f}M "
        f"(max_persistent: {max_persistent_str})"
    )



import math
import torch
from torch import amp
import torch.nn as nn
import torch.nn.functional as F

from einops import rearrange
from diffusers import ModelMixin
from diffusers.configuration_utils import ConfigMixin, register_to_config

from ..layers.rope import VideoRopePosition3DEmb
from ..layers.attention import SingleStreamMutiAttention
from ..kernels.rope import fast_rope_apply, sinusoidal_embedding_1d
from ..kernels.attn import attention
from ..utils import get_attn_map_with_target


class WanRMSNorm(nn.Module):
    def __init__(self, dim, eps=1e-5):
        super().__init__()
        self.dim = dim
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def forward(self, x):
        r"""
        Args:
            x(Tensor): Shape [B, L, C]
        """
        return self._norm(x.float()).type_as(x) * self.weight

    def _norm(self, x):
        return x * torch.rsqrt(x.pow(2).mean(dim=-1, keepdim=True) + self.eps)


class WanLayerNorm(nn.LayerNorm):
    def __init__(self, dim, eps=1e-6, elementwise_affine=False):
        super().__init__(dim, elementwise_affine=elementwise_affine, eps=eps)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        origin_dtype = inputs.dtype
        out = F.layer_norm(
            inputs.float(),
            self.normalized_shape,
            None if self.weight is None else self.weight.float(),
            None if self.bias is None else self.bias.float(),
            self.eps,
        ).to(origin_dtype)
        return out


class WanSelfAttention(nn.Module):
    def __init__(self, dim, num_heads, window_size=(-1, -1), qk_norm=True, eps=1e-6):
        assert dim % num_heads == 0
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.window_size = window_size
        self.qk_norm = qk_norm
        self.eps = eps

        # layers
        self.q = nn.Linear(dim, dim)
        self.k = nn.Linear(dim, dim)
        self.v = nn.Linear(dim, dim)
        self.o = nn.Linear(dim, dim)
        self.norm_q = WanRMSNorm(dim, eps=eps) if qk_norm else nn.Identity()
        self.norm_k = WanRMSNorm(dim, eps=eps) if qk_norm else nn.Identity()

    def forward(self, x, seq_lens, grid_sizes, freqs, ref_target_masks=None):
        b, s, n, d = *x.shape[:2], self.num_heads, self.head_dim

        # query, key, value function
        def qkv_fn(x):
            q = self.norm_q(self.q(x)).view(b, s, n, d)
            k = self.norm_k(self.k(x)).view(b, s, n, d)
            v = self.v(x).view(b, s, n, d)
            return q, k, v

        q, k, v = qkv_fn(x)
        q = fast_rope_apply(q, freqs)
        k = fast_rope_apply(k, freqs)

        x = attention(q=q, k=k, v=v).type_as(x)

        # output
        x = x.flatten(2)
        x = self.o(x)

        if ref_target_masks is not None:
            with torch.no_grad():
                x_ref_attn_map = get_attn_map_with_target(
                    q.type_as(x),
                    k.type_as(x),
                    grid_sizes[0],
                    ref_target_masks=ref_target_masks,
                )
        else:
            x_ref_attn_map = None

        return x, x_ref_attn_map


class WanI2VCrossAttention(WanSelfAttention):
    def __init__(self, dim, num_heads, window_size=(-1, -1), qk_norm=True, eps=1e-6):
        super().__init__(dim, num_heads, window_size, qk_norm, eps)

        self.k_img = nn.Linear(dim, dim)
        self.v_img = nn.Linear(dim, dim)
        self.norm_k_img = WanRMSNorm(dim, eps=eps) if qk_norm else nn.Identity()

    def forward(self, x, context, context_lens):
        context_img = context[:, :257]
        context = context[:, 257:]
        b, n, d = x.size(0), self.num_heads, self.head_dim

        # compute query, key, value
        q = self.norm_q(self.q(x)).view(b, -1, n, d)
        k = self.norm_k(self.k(context)).view(b, -1, n, d)
        v = self.v(context).view(b, -1, n, d)
        k_img = self.norm_k_img(self.k_img(context_img)).view(b, -1, n, d)
        v_img = self.v_img(context_img).view(b, -1, n, d)
        img_x = attention(q, k_img, v_img)
        # compute attention
        x = attention(q, k, v)

        # output
        x = x.flatten(2)
        img_x = img_x.flatten(2)
        x = x + img_x
        x = self.o(x)
        return x


class WanAttentionBlock(nn.Module):
    def __init__(
        self,
        cross_attn_type,
        dim,
        ffn_dim,
        num_heads,
        window_size=(-1, -1),
        qk_norm=True,
        cross_attn_norm=False,
        eps=1e-6,
        output_dim=768,
        norm_input_visual=True,
        class_range=24,
        class_interval=4,
    ):
        super().__init__()
        self.dim = dim
        self.ffn_dim = ffn_dim
        self.num_heads = num_heads
        self.window_size = window_size
        self.qk_norm = qk_norm
        self.cross_attn_norm = cross_attn_norm
        self.eps = eps

        # layers
        self.norm1 = WanLayerNorm(dim, eps)
        self.self_attn = WanSelfAttention(dim, num_heads, window_size, qk_norm, eps)
        self.norm3 = (
            WanLayerNorm(dim, eps, elementwise_affine=True)
            if cross_attn_norm
            else nn.Identity()
        )
        self.cross_attn = WanI2VCrossAttention(dim, num_heads, (-1, -1), qk_norm, eps)
        self.norm2 = WanLayerNorm(dim, eps)
        self.ffn = nn.Sequential(
            nn.Linear(dim, ffn_dim),
            nn.GELU(approximate="tanh"),
            nn.Linear(ffn_dim, dim),
        )

        # modulation
        self.modulation = nn.Parameter(torch.randn(1, 6, dim) / dim**0.5)

        # init audio module
        self.audio_cross_attn = SingleStreamMutiAttention(
            dim=dim,
            encoder_hidden_states_dim=output_dim,
            num_heads=num_heads,
            qk_norm=False,
            qkv_bias=True,
            eps=eps,
            norm_layer=WanRMSNorm,
            class_range=class_range,
            class_interval=class_interval,
        )
        self.norm_x = (
            WanLayerNorm(dim, eps, elementwise_affine=True)
            if norm_input_visual
            else nn.Identity()
        )

    def forward(
        self,
        x,
        e,
        seq_lens,
        grid_sizes,
        freqs,
        context,
        context_lens,
        audio_embedding=None,
        ref_target_masks=None,
        human_num=None,
    ):

        dtype = x.dtype
        assert e.dtype == torch.float32
        with amp.autocast("cuda", dtype=torch.float32):
            e = (self.modulation.to(e.device) + e).chunk(6, dim=1)
        assert e[0].dtype == torch.float32

        # self-attention
        y, x_ref_attn_map = self.self_attn(
            (self.norm1(x).float() * (1 + e[1]) + e[0]).type_as(x),
            seq_lens,
            grid_sizes,
            freqs,
            ref_target_masks=ref_target_masks,
        )
        with amp.autocast("cuda", dtype=torch.float32):
            x = x + y * e[2]

        x = x.to(dtype)

        # cross-attention of text
        x = x + self.cross_attn(self.norm3(x), context, context_lens)

        # cross attn of audio
        x_a = self.audio_cross_attn(
            self.norm_x(x),
            encoder_hidden_states=audio_embedding,
            shape=grid_sizes[0],
            x_ref_attn_map=x_ref_attn_map,
            human_num=human_num,
        )
        x = x + x_a

        y = self.ffn((self.norm2(x).float() * (1 + e[4]) + e[3]).to(dtype))
        with amp.autocast("cuda", dtype=torch.float32):
            x = x + y * e[5]

        x = x.to(dtype)

        return x


class Head(nn.Module):
    def __init__(self, dim, out_dim, patch_size, eps=1e-6):
        super().__init__()
        self.dim = dim
        self.out_dim = out_dim
        self.patch_size = patch_size
        self.eps = eps

        # layers
        out_dim = math.prod(patch_size) * out_dim
        self.norm = WanLayerNorm(dim, eps)
        self.head = nn.Linear(dim, out_dim)

        # modulation
        self.modulation = nn.Parameter(torch.randn(1, 2, dim) / dim**0.5)

    def forward(self, x, e):
        r"""
        Args:
            x(Tensor): Shape [B, L1, C]
            e(Tensor): Shape [B, C]
        """
        assert e.dtype == torch.float32
        with amp.autocast("cuda", dtype=torch.float32):
            e = (self.modulation.to(e.device) + e.unsqueeze(1)).chunk(2, dim=1)
            x = self.head(self.norm(x) * (1 + e[1]) + e[0])
        return x


class MLPProj(torch.nn.Module):
    def __init__(self, in_dim, out_dim):
        super().__init__()

        self.proj = torch.nn.Sequential(
            torch.nn.LayerNorm(in_dim),
            torch.nn.Linear(in_dim, in_dim),
            torch.nn.GELU(),
            torch.nn.Linear(in_dim, out_dim),
            torch.nn.LayerNorm(out_dim),
        )

    def forward(self, image_embeds):
        clip_extra_context_tokens = self.proj(image_embeds)
        return clip_extra_context_tokens


class AudioProjModel(ModelMixin, ConfigMixin):
    def __init__(
        self,
        seq_len=5,
        seq_len_vf=12,
        blocks=12,
        channels=768,
        intermediate_dim=512,
        output_dim=768,
        context_tokens=32,
        norm_output_audio=False,
    ):
        super().__init__()

        self.seq_len = seq_len
        self.blocks = blocks
        self.channels = channels
        self.input_dim = seq_len * blocks * channels
        self.input_dim_vf = seq_len_vf * blocks * channels
        self.intermediate_dim = intermediate_dim
        self.context_tokens = context_tokens
        self.output_dim = output_dim

        # define multiple linear layers
        self.proj1 = nn.Linear(self.input_dim, intermediate_dim)
        self.proj1_vf = nn.Linear(self.input_dim_vf, intermediate_dim)
        self.proj2 = nn.Linear(intermediate_dim, intermediate_dim)
        self.proj3 = nn.Linear(intermediate_dim, context_tokens * output_dim)
        self.norm = nn.LayerNorm(output_dim) if norm_output_audio else nn.Identity()

    def forward(self, audio_embeds, audio_embeds_vf):
        video_length = audio_embeds.shape[1] + audio_embeds_vf.shape[1]
        B, _, _, S, C = audio_embeds.shape

        # process audio of first frame
        audio_embeds = rearrange(audio_embeds, "bz f w b c -> (bz f) w b c")
        batch_size, window_size, blocks, channels = audio_embeds.shape
        audio_embeds = audio_embeds.view(batch_size, window_size * blocks * channels)

        # process audio of latter frame
        audio_embeds_vf = rearrange(audio_embeds_vf, "bz f w b c -> (bz f) w b c")
        batch_size_vf, window_size_vf, blocks_vf, channels_vf = audio_embeds_vf.shape
        audio_embeds_vf = audio_embeds_vf.view(
            batch_size_vf, window_size_vf * blocks_vf * channels_vf
        )

        # first projection
        audio_embeds = torch.relu(self.proj1(audio_embeds))
        audio_embeds_vf = torch.relu(self.proj1_vf(audio_embeds_vf))
        audio_embeds = rearrange(audio_embeds, "(bz f) c -> bz f c", bz=B)
        audio_embeds_vf = rearrange(audio_embeds_vf, "(bz f) c -> bz f c", bz=B)
        audio_embeds_c = torch.concat([audio_embeds, audio_embeds_vf], dim=1)
        batch_size_c, N_t, C_a = audio_embeds_c.shape
        audio_embeds_c = audio_embeds_c.view(batch_size_c * N_t, C_a)

        # second projection
        audio_embeds_c = torch.relu(self.proj2(audio_embeds_c))

        context_tokens = self.proj3(audio_embeds_c).reshape(
            batch_size_c * N_t, self.context_tokens, self.output_dim
        )

        # normalization and reshape
        with amp.autocast("cuda", dtype=torch.float32):
            context_tokens = self.norm(context_tokens)
        context_tokens = rearrange(
            context_tokens, "(bz f) m c -> bz f m c", f=video_length
        )

        return context_tokens


class WanModel(ModelMixin, ConfigMixin):
    r"""
    Wan diffusion backbone supporting both text-to-video and image-to-video.
    """

    ignore_for_config = [
        "patch_size",
        "cross_attn_norm",
        "qk_norm",
        "text_dim",
        "window_size",
    ]
    _no_split_modules = ["WanAttentionBlock"]

    @register_to_config
    def __init__(
        self,
        model_type="i2v",
        patch_size=(1, 2, 2),
        text_len=512,
        in_dim=16,
        dim=2048,
        ffn_dim=8192,
        freq_dim=256,
        text_dim=4096,
        out_dim=16,
        num_heads=16,
        num_layers=32,
        window_size=(-1, -1),
        qk_norm=True,
        cross_attn_norm=True,
        eps=1e-6,
        # audio params
        audio_window=5,
        intermediate_dim=512,
        output_dim=768,
        context_tokens=32,
        vae_scale=4,  # vae timedownsample scale
        norm_input_visual=True,
        norm_output_audio=True,
        weight_init=True,
    ):
        super().__init__()

        assert model_type == "i2v", "MultiTalk model requires your model_type is i2v."
        self.model_type = model_type

        self.patch_size = patch_size
        self.text_len = text_len
        self.in_dim = in_dim
        self.dim = dim
        self.ffn_dim = ffn_dim
        self.freq_dim = freq_dim
        self.text_dim = text_dim
        self.out_dim = out_dim
        self.num_heads = num_heads
        self.num_layers = num_layers
        self.window_size = window_size
        self.qk_norm = qk_norm
        self.cross_attn_norm = cross_attn_norm
        self.eps = eps

        self.norm_output_audio = norm_output_audio
        self.audio_window = audio_window
        self.intermediate_dim = intermediate_dim
        self.vae_scale = vae_scale

        # embeddings
        self.patch_embedding = nn.Conv3d(
            in_dim, dim, kernel_size=patch_size, stride=patch_size
        )
        self.text_embedding = nn.Sequential(
            nn.Linear(text_dim, dim), nn.GELU(approximate="tanh"), nn.Linear(dim, dim)
        )

        self.time_embedding = nn.Sequential(
            nn.Linear(freq_dim, dim), nn.SiLU(), nn.Linear(dim, dim)
        )
        self.time_projection = nn.Sequential(nn.SiLU(), nn.Linear(dim, dim * 6))

        # blocks
        cross_attn_type = "i2v_cross_attn"
        self.blocks = nn.ModuleList(
            [
                WanAttentionBlock(
                    cross_attn_type,
                    dim,
                    ffn_dim,
                    num_heads,
                    window_size,
                    qk_norm,
                    cross_attn_norm,
                    eps,
                    output_dim=output_dim,
                    norm_input_visual=norm_input_visual,
                )
                for _ in range(num_layers)
            ]
        )

        # head
        self.head = Head(dim, out_dim, patch_size, eps)

        assert (dim % num_heads) == 0 and (dim // num_heads) % 2 == 0
        self.head_dim = dim // num_heads
        self.rope = VideoRopePosition3DEmb(
            head_dim=self.head_dim, len_h=1024, len_w=1024, len_t=1024
        )

        if model_type == "i2v":
            self.img_emb = MLPProj(1280, dim)
        else:
            raise NotImplementedError("Not supported model type.")

        # init audio adapter
        self.audio_proj = AudioProjModel(
            seq_len=audio_window,
            seq_len_vf=audio_window + vae_scale - 1,
            intermediate_dim=intermediate_dim,
            output_dim=output_dim,
            context_tokens=context_tokens,
            norm_output_audio=norm_output_audio,
        )

        # initialize weights
        if weight_init:
            self.init_weights()

    def forward(
        self,
        x,
        t,
        context,
        seq_len,
        clip_fea=None,
        y=None,
        audio=None,
        ref_target_masks=None,
    ):
        assert clip_fea is not None and y is not None

        # params
        # device = self.patch_embedding.weight.device
        # if self.freqs.device != device:
        #     self.freqs = self.freqs.to(device)
        B = 1
        _, T, H, W = x[0].shape
        N_t = T // self.patch_size[0]
        N_h = H // self.patch_size[1]
        N_w = W // self.patch_size[2]

        freqs_shape = torch.Size([B, T, H // 2, W // 2, self.head_dim])
        self.freqs = self.rope.generate_embeddings(B_T_H_W_C=freqs_shape)

        if y is not None:
            x = [torch.cat([u, v], dim=0) for u, v in zip(x, y)]
        x[0] = x[0].to(context[0].dtype)

        # embeddings
        x = [self.patch_embedding(u.unsqueeze(0)) for u in x]
        grid_sizes = torch.stack(
            [torch.tensor(u.shape[2:], dtype=torch.long) for u in x]
        )
        x = [u.flatten(2).transpose(1, 2) for u in x]  # [1 fhw c]

        seq_lens = torch.tensor([u.size(1) for u in x], dtype=torch.long)
        assert seq_lens.max() <= seq_len
        x = torch.cat(
            [
                torch.cat([u, u.new_zeros(1, seq_len - u.size(1), u.size(2))], dim=1)
                for u in x
            ]
        )

        # time embeddings
        with amp.autocast("cuda", dtype=torch.float32):
            e = self.time_embedding(sinusoidal_embedding_1d(self.freq_dim, t).float())
            e0 = self.time_projection(e).unflatten(1, (6, self.dim))
            assert e.dtype == torch.float32 and e0.dtype == torch.float32

        # text embedding
        context_lens = None
        context = self.text_embedding(
            torch.stack(
                [
                    torch.cat([u, u.new_zeros(self.text_len - u.size(0), u.size(1))])
                    for u in context
                ]
            )
        )

        # clip embedding
        if clip_fea is not None:
            context_clip = self.img_emb(clip_fea)
            context = torch.concat([context_clip, context], dim=1).to(x.dtype)

        audio_cond = audio.to(device=x.device, dtype=x.dtype)
        first_frame_audio_emb_s = audio_cond[:, :1, ...]  # b 1 w s c
        latter_frame_audio_emb = audio_cond[:, 1:, ...]
        latter_frame_audio_emb = rearrange(
            latter_frame_audio_emb, "b (n_t n) w s c -> b n_t n w s c", n=self.vae_scale
        )
        middle_index = self.audio_window // 2
        latter_first_frame_audio_emb = latter_frame_audio_emb[
            :, :, :1, : middle_index + 1, ...
        ]
        latter_first_frame_audio_emb = rearrange(
            latter_first_frame_audio_emb, "b n_t n w s c -> b n_t (n w) s c"
        )  # b n_t (1 3) s c
        latter_last_frame_audio_emb = latter_frame_audio_emb[
            :, :, -1:, middle_index:, ...
        ]
        latter_last_frame_audio_emb = rearrange(
            latter_last_frame_audio_emb, "b n_t n w s c -> b n_t (n w) s c"
        )  # b n_t (1 3) s c
        latter_middle_frame_audio_emb = latter_frame_audio_emb[
            :, :, 1:-1, middle_index : middle_index + 1, ...
        ]
        latter_middle_frame_audio_emb = rearrange(
            latter_middle_frame_audio_emb, "b n_t n w s c -> b n_t (n w) s c"
        )  # b n_t (2 1) s c

        latter_frame_audio_emb_s = torch.concat(
            [
                latter_first_frame_audio_emb,
                latter_middle_frame_audio_emb,
                latter_last_frame_audio_emb,
            ],
            dim=2,
        )
        audio_embedding = self.audio_proj(
            first_frame_audio_emb_s, latter_frame_audio_emb_s
        )
        human_num = len(audio_embedding)
        # if False:
        #     audio_embedding = torch.concat(audio_embedding.split(1), dim=2).to(x.dtype)
        # else:
        #     audio_embedding = audio_embedding.to(x.dtype)
        audio_embedding = torch.concat(audio_embedding.split(1), dim=2).to(x.dtype)

        # convert ref_target_masks to token_ref_target_masks
        if ref_target_masks is not None:
            ref_target_masks = ref_target_masks.unsqueeze(0).to(torch.float32)
            token_ref_target_masks = nn.functional.interpolate(
                ref_target_masks, size=(N_h, N_w), mode="nearest"
            )
            token_ref_target_masks = token_ref_target_masks.squeeze(0)
            token_ref_target_masks = token_ref_target_masks > 0
            token_ref_target_masks = token_ref_target_masks.view(
                token_ref_target_masks.shape[0], -1
            )
            token_ref_target_masks = token_ref_target_masks.to(x.dtype)
        else:
            token_ref_target_masks = None

        # arguments
        kwargs = dict(
            e=e0,
            seq_lens=seq_lens,
            grid_sizes=grid_sizes,
            freqs=self.freqs,
            context=context,
            context_lens=context_lens,
            audio_embedding=audio_embedding,
            ref_target_masks=token_ref_target_masks,
            human_num=human_num,
        )
        for block in self.blocks:
            x = block(x, **kwargs)

        # head
        x = self.head(x, e)

        # unpatchify
        x = self.unpatchify(x, grid_sizes)

        return torch.stack(x).float()

    def unpatchify(self, x, grid_sizes):
        r"""
        Reconstruct video tensors from patch embeddings.

        Args:
            x (List[Tensor]):
                List of patchified features, each with shape [L, C_out * prod(patch_size)]
            grid_sizes (Tensor):
                Original spatial-temporal grid dimensions before patching,
                    shape [B, 3] (3 dimensions correspond to F_patches, H_patches, W_patches)

        Returns:
            List[Tensor]:
                Reconstructed video tensors with shape [C_out, F, H / 8, W / 8]
        """

        c = self.out_dim
        out = []
        for u, v in zip(x, grid_sizes.tolist()):
            u = u[: math.prod(v)].view(*v, *self.patch_size, c)
            u = torch.einsum("fhwpqrc->cfphqwr", u)
            u = u.reshape(c, *[i * j for i, j in zip(v, self.patch_size)])
            out.append(u)
        return out

    def init_weights(self):
        r"""
        Initialize model parameters using Xavier initialization.
        """

        # basic init
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

        # init embeddings
        nn.init.xavier_uniform_(self.patch_embedding.weight.flatten(1))
        for m in self.text_embedding.modules():
            if isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, std=0.02)
        for m in self.time_embedding.modules():
            if isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, std=0.02)

        # init output layer
        nn.init.zeros_(self.head.head.weight)



import torch
import math
import triton
import triton.language as tl
##############################################################################################


# code based https://github.com/fpgaminer/GPTQ-triton
def kernel_config_pruner(configs, nargs, **kwargs):
    m = max(
        2 ** int(math.ceil(math.log2(nargs["M"]))), 16
    )  # Need at least 16 here for tl.dot
    n = max(2 ** int(math.ceil(math.log2(nargs["N"]))), 16)
    k = max(2 ** int(math.ceil(math.log2(nargs["K"]))), 16)
    g = nargs["group_size"]

    used = set()
    for config in configs:
        group_size_m = config.kwargs["GROUP_SIZE_M"]
        block_size_m = min(m, config.kwargs["BLOCK_SIZE_M"])
        block_size_n = min(n, config.kwargs["BLOCK_SIZE_N"])
        block_size_k = min(k, config.kwargs["BLOCK_SIZE_K"])
        block_size_k = min(
            block_size_k, g
        )  # Makes BLOCK_SIZE_K compatible with the group_size

        if (
            block_size_m,
            block_size_n,
            block_size_k,
            group_size_m,
            config.num_stages,
            config.num_warps,
        ) in used:
            continue

        used.add(
            (
                block_size_m,
                block_size_n,
                block_size_k,
                group_size_m,
                config.num_stages,
                config.num_warps,
            )
        )
        yield triton.Config(
            {
                "BLOCK_SIZE_M": block_size_m,
                "BLOCK_SIZE_N": block_size_n,
                "BLOCK_SIZE_K": block_size_k,
                "GROUP_SIZE_M": group_size_m,
            },
            num_stages=config.num_stages,
            num_warps=config.num_warps,
        )


def get_gemm_config():
    # Tuned on 4090 RTX
    _configs = []
    for _M in [16, 32, 64, 128]:  # might need higher values for larger batch-sizes
        for _N in [32, 64, 128]:
            for _K in [32, 64, 128]:  # [32, 64, 128], 32 <= block_size
                for _w in [2, 4]:
                    for _s in [2, 4]:
                        _configs.append(
                            triton.Config(
                                {
                                    "BLOCK_SIZE_M": _M,
                                    "BLOCK_SIZE_N": _N,
                                    "BLOCK_SIZE_K": _K,
                                    "GROUP_SIZE_M": 8,
                                },
                                num_stages=_s,
                                num_warps=_w,
                            )
                        )
    return _configs


@triton.autotune(
    configs=get_gemm_config(),
    key=["M", "N", "K", "group_size", "W_nbits_1", "W_nbits_2"],
    prune_configs_by={
        "early_config_prune": kernel_config_pruner,
    },
    warmup=200,
    rep=50,  # 20 for faster tuning
)
@triton.jit
def gemm_A16fWnO16f_int32packing_kernel(
    a_ptr,
    b_1_ptr,
    b_2_ptr,
    c_ptr,
    scales_ptr,
    zeros_ptr,
    M,
    N,
    K,
    W_nbits_1: tl.constexpr,
    W_nbits_2: tl.constexpr,
    group_size: tl.constexpr,
    unpack_mask_1: tl.constexpr,
    unpack_mask_2: tl.constexpr,
    elements_per_sample_1: tl.constexpr,
    elements_per_sample_2: tl.constexpr,
    stride_am,
    stride_ak,
    stride_bk_1,
    stride_bn_1,
    stride_bk_2,
    stride_bn_2,
    stride_cm,
    stride_cn,
    stride_meta,
    acc_dtype: tl.constexpr,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    BLOCK_SIZE_K: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
    num_stages: tl.constexpr,
):
    """
    Based on https://github.com/fpgaminer/GPTQ-triton
    GEMM for C = matmul(A, dequantize(B, scales, zeros))
    A is of shape (M, K): float16 or bfloat16
    B is of shape (K//8, N): int32 as a packed matrix
    C is of shape (M, N): float16 or bfloat16 depending on the input A
    scales and zeros is of shape (group_size, N): float16 or bfloat16

    BLOCK_SIZE_M >=16
    BLOCK_SIZE_K <= group_size
    """

    pid = tl.program_id(axis=0)
    num_pid_m = tl.cdiv(M, BLOCK_SIZE_M)
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)
    num_pid_k = tl.cdiv(K, BLOCK_SIZE_K)

    num_pid_in_group = GROUP_SIZE_M * num_pid_n
    group_id = pid // num_pid_in_group
    first_pid_m = group_id * GROUP_SIZE_M
    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)
    pid_m = first_pid_m + (pid % group_size_m)
    pid_n = (pid % num_pid_in_group) // group_size_m

    # Offsets
    offs_am = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_bn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    offs_k = tl.arange(0, BLOCK_SIZE_K)

    # Inputs
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)
    a_mask = offs_am[:, None] < M
    b_ptrs_1 = b_1_ptr + (
        (offs_k[:, None] // elements_per_sample_1) * stride_bk_1
        + offs_bn[None, :] * stride_bn_1
    )
    b_ptrs_2 = b_2_ptr + (
        (offs_k[:, None] // elements_per_sample_2) * stride_bk_2
        + offs_bn[None, :] * stride_bn_2
    )

    # Output
    c_ptrs = c_ptr + stride_cm * offs_am[:, None] + stride_cn * offs_bn[None, :]

    # Meta data stuff
    q_shifts1 = ((offs_k % elements_per_sample_1) * W_nbits_1).to(tl.int32)[:, None]
    q_shifts2 = ((offs_k % elements_per_sample_2) * W_nbits_2).to(tl.int32)[:, None]
    scales_ptrs = scales_ptr + offs_bn[None, :]
    zeros_ptrs = zeros_ptr + offs_bn[None, :]
    stride_mul = BLOCK_SIZE_K / group_size

    ####################################################################################
    acc = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=acc_dtype)

    for k in tl.range(0, num_pid_k, 1, num_stages=1):
        b1 = (
            (tl.load(b_ptrs_1, eviction_policy="evict_first") >> q_shifts1)
            & unpack_mask_1
        ).to(tl.uint8)  # (BLOCK_SIZE_K, BLOCK_SIZE_N) - repeated over K dim
        b2 = (
            (tl.load(b_ptrs_2, eviction_policy="evict_first") >> q_shifts2)
            & unpack_mask_2
        ).to(tl.uint8)  # (BLOCK_SIZE_K, BLOCK_SIZE_N) - repeated over K dim

        k_m = (k * stride_mul).to(tl.int32)
        scales = tl.load(scales_ptrs + k_m * stride_meta)
        zeros = tl.load(zeros_ptrs + k_m * stride_meta)

        a = tl.load(
            a_ptrs, mask=a_mask, other=0.0, eviction_policy="evict_last"
        )  # (BLOCK_SIZE_M, BLOCK_SIZE_K)

        # Unpack and dequantize
        b = (((b2 << 1) | b1).to(a.dtype) - zeros) * scales

        # Dot
        acc = tl.dot(
            a, b, acc=acc, out_dtype=acc_dtype, input_precision="ieee"
        )  # (BLOCK_SIZE_M, BLOCK_SIZE_N)

        # Advance
        a_ptrs += BLOCK_SIZE_K
        b_ptrs_1 += (BLOCK_SIZE_K // elements_per_sample_1) * stride_bk_1
        b_ptrs_2 += (BLOCK_SIZE_K // elements_per_sample_2) * stride_bk_2

    tl.store(c_ptrs, acc, mask=(offs_am[:, None] < M) & (offs_bn[None, :] < N))


def gemm_A16fWnO16f_int32packing_forward(
    x,
    W_q_1bit,
    W_q_2bit,
    scales,
    zeros,
    W_nbits_1,
    W_nbits_2,
    group_size,
    unpack_mask_1,
    unpack_mask_2,
    elements_per_sample_1,
    elements_per_sample_2,
    acc_dtype=tl.float16,
):
    output = torch.empty(
        (x.shape[0], W_q_1bit.shape[1]), device=W_q_1bit.device, dtype=scales.dtype
    )

    # assert x.shape[1] == W_q.shape[0] * elements_per_sample, "Invalid Input Shapes"

    grid = lambda META: (
        triton.cdiv(x.shape[0], META["BLOCK_SIZE_M"])
        * triton.cdiv(W_q_1bit.shape[1], META["BLOCK_SIZE_N"]),
    )

    gemm_A16fWnO16f_int32packing_kernel[grid](
        x,
        W_q_1bit,
        W_q_2bit,
        output,
        scales,
        zeros,
        x.shape[0],
        W_q_1bit.shape[1],
        x.shape[1],
        W_nbits_1,
        W_nbits_2,
        group_size,
        unpack_mask_1,
        unpack_mask_2,
        elements_per_sample_1,
        elements_per_sample_2,
        x.stride(0),
        x.stride(1),
        W_q_1bit.stride(0),
        W_q_1bit.stride(1),
        W_q_2bit.stride(0),
        W_q_2bit.stride(1),
        output.stride(0),
        output.stride(1),
        scales.stride(0),
        acc_dtype,
    )

    return output


class GemliteLinear(torch.nn.Module):
    def __init__(self, W_nbits, group_size, in_features, out_features, bias):

        self.SUPPORTED_BITS = [5]  # + [3]

        super().__init__()
        if W_nbits not in self.SUPPORTED_BITS:
            raise NotImplementedError("Only 5 W_nbits are supported.")
        # if in_features % 256 != 0 or out_features % 256 != 0:
        #     raise NotImplementedError("in_feature or out_feature must be divisible by 256.")
        self.in_features = in_features
        self.out_features = out_features
        self.orig_shape = (out_features, in_features)
        self.group_size = group_size if group_size != -1 else in_features

        self.W_nbits_1 = 1
        self.unpack_mask_1 = 2**self.W_nbits_1 - 1
        self.elements_per_sample_1 = 32 // self.W_nbits_1

        self.W_nbits_2 = 4
        self.unpack_mask_2 = 2**self.W_nbits_2 - 1
        self.elements_per_sample_2 = 32 // self.W_nbits_2

        self.bias = None

    def pack_base(self, W_q, n_bits):
        W_q = W_q.reshape(self.orig_shape).t().contiguous().to(torch.int32)
        W_q_out = torch.zeros(
            (W_q.shape[0] // 32 * n_bits, W_q.shape[1]),
            dtype=torch.int32,
            device=W_q.device,
        )  # Packed

        step = 32 // n_bits
        i, row = 0, 0
        while row < W_q_out.shape[0]:
            shift = 0
            for j in range(i, i + step):
                W_q_out[row] |= W_q[j] << shift
                shift += n_bits
            i += step
            row += 1

        return W_q_out.contiguous()

    def pack(self, W_q, scales, zeros):
        self.W_q_1bit = self.pack_base(W_q.int() & 0b1, self.W_nbits_1)
        self.W_q_2bit = self.pack_base((W_q.int() >> 1) & 0b1111, self.W_nbits_2)
        self.scales = scales.reshape((self.out_features, -1)).t().contiguous()
        self.zeros = zeros.reshape((self.out_features, -1)).t().contiguous()
        return self

    def forward(self, x):
        out_shape = x.shape[:-1] + (self.out_features,)

        out = gemm_A16fWnO16f_int32packing_forward(
            x.reshape(-1, x.shape[-1]),
            self.W_q_1bit,
            self.W_q_2bit,
            self.scales,
            self.zeros,
            self.W_nbits_1,
            self.W_nbits_2,
            self.group_size,
            self.unpack_mask_1,
            self.unpack_mask_2,
            self.elements_per_sample_1,
            self.elements_per_sample_2,
        ).reshape(out_shape)

        if self.bias is not None:
            out += self.bias
        return out


#################################################################################################################################
torch._dynamo.config.capture_scalar_outputs = True
torch._inductor.config.coordinate_descent_tuning = True


@torch.compile()
def matmul_torch_A16W8SYM(x, W_q, scales, out_features):
    out_shape = x.shape[:-1] + (out_features,)
    out = ((x.view((-1, x.shape[-1])) @ W_q.T.to(x.dtype)) / scales.view(1, -1)).view(
        out_shape
    )
    return out


class Torch_A16W8SYM(torch.nn.Module):
    def __init__(self, in_features, out_features, W_q, scales, bias=None):
        super().__init__()
        self.W_q = W_q
        self.in_features = in_features
        self.out_features = out_features
        self.scales = scales
        self.bias = bias
        self.device = W_q.device
        self.compute_dtype = scales.dtype

    def forward(self, x):
        out = matmul_torch_A16W8SYM(
            x.to(self.device), self.W_q, self.scales, self.out_features
        )
        if self.bias is not None:
            out += self.bias
        return out


from hqq.core.quantize import *
from hqq.backends.bitblas import HQQLinearBitBlas


import numpy as np
from triton.testing import do_bench


def eval_time(fct, params):
    return do_bench(
        lambda: fct(**params), warmup=25, rep=200, fast_flush=True, return_mode="min"
    )


torch.random.manual_seed(100)


class empty_linear(torch.nn.Module):
    def __init__(self, in_features, out_features, compute_dtype, device):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.device = device
        self.compute_dtype = compute_dtype


def gen_data(in_features, out_features, W_nbits, group_size, device="cuda:0"):
    linear = torch.nn.Linear(
        in_features=in_features, out_features=out_features, bias=False, device="cpu"
    )

    compute_dtype = torch.float16

    quant_config = BaseQuantizeConfig(
        nbits=W_nbits,
        group_size=group_size,
        quant_zero=False,
        quant_scale=False,
        axis=1,
    )
    # hqq_layer    = HQQLinear(linear, quant_config=quant_config, compute_dtype=compute_dtype, device=device, del_orig=False) #bfloat16

    orig_shape = (out_features, in_features)
    # W            = hqq_layer.dequantize().reshape(orig_shape)

    triton_linear, torchao_linear, bitblas_linear, marlin_linear = [None] * 4

    W_q = torch.randint(
        0, 2**W_nbits, (out_features, in_features), dtype=torch.uint8, device=device
    ).to(torch.uint8)

    orig_shape = (out_features, in_features)
    N = in_features * out_features // group_size

    # #scales = torch.ones((N,), dtype=compute_dtype, device='cuda:0')
    scales = torch.randn((N,), dtype=compute_dtype, device=device).abs() / 500.0

    # #zeros  = torch.zeros((N,), dtype=compute_dtype, device='cuda:0')
    zeros = torch.randint(0, 2**W_nbits - 1, (N,), dtype=compute_dtype, device=device)

    W = (
        (W_q.reshape([-1, group_size]) - zeros.view((N, 1))) * scales.view((N, 1))
    ).reshape(orig_shape)

    # Triton
    if W_nbits in [5]:
        # Z             = W.numel() // group_size
        # W_q           = hqq_layer.unpack()[:Z,:].reshape(orig_shape).clone()
        # scales        = hqq_layer.meta['scale'].clone()
        # zeros         = hqq_layer.meta['zero'].clone()
        triton_linear = GemliteLinear(
            W_nbits=W_nbits,
            group_size=group_size,
            in_features=in_features,
            out_features=out_features,
            bias=False,
        ).pack(W_q, scales, zeros)
        triton_linear.compute_dtype = scales.dtype

    # #TorchAO
    # if(W_nbits==8):
    #     torchao_linear = Torch_A16W8SYM(in_features, out_features, (W_q.int() - 127).to(torch.int8), scales, bias=None)
    # if(W_nbits==4):
    #     torchao_linear = patch_hqq_to_aoint4(hqq_layer, None)

    # # Bitblas
    # if(W_nbits in [8, 4, 2]):
    #     bitblas_linear = patch_hqq_to_bitblas(HQQLinear(linear, quant_config=quant_config, compute_dtype=torch.float16, device=device, del_orig=False), None)

    #################################################################
    # #Marlin
    # from vllm.model_executor.layers.quantization.awq_marlin import AWQMarlinLinearMethod as MarlinLinearMethod
    # from vllm.model_executor.layers.quantization.awq_marlin import AWQMarlinConfig as MarlinConfig

    # if(W_nbits==4):
    #     _marlin_linear = MarlinLinearMethod(MarlinConfig(weight_bits=W_nbits, group_size=group_size, has_zp=True, lm_head_quantized=False))

    #     marlin_linear = empty_linear(in_features, out_features, compute_dtype=torch.float16, device='cuda:0')
    #     _marlin_linear.create_weights(layer=marlin_linear,
    #             input_size_per_partition=in_features,
    #             output_partition_sizes=[out_features],
    #             input_size=in_features,
    #             output_size=out_features,
    #             params_dtype=torch.float16)

    #     marlin_linear = marlin_linear.cuda()
    #     _marlin_linear.process_weights_after_loading(marlin_linear)

    #     marlin_linear.scales.data = torch.zeros_like(marlin_linear.scales.data) + 1
    #     marlin_linear.bias = None
    #     marlin_linear.forward = lambda x: _marlin_linear.apply(layer=marlin_linear, x=x, bias=marlin_linear.bias)
    # #################################################################

    del linear
    torch.cuda.empty_cache()

    return W, triton_linear, torchao_linear, bitblas_linear, marlin_linear


def check_valid(x, W, quant_linear):

    y_ref = torch.matmul(x, W.T)
    y_q = quant_linear(x)

    # print("ref", y_ref)
    # print("quantized", y_q)
    # print("max error", (y_ref - y_q).abs().max())

    try:
        assert (y_ref - y_q).abs().mean() < 5e-3
    except:
        raise Error("assertion failed")


#############################################################################################################
# in_features, out_features = 4096, 4096
# in_features, out_features = 4096*2, 4096*2
# in_features, out_features = 4096*4, 4096*4
in_features, out_features = 4096 * 8, 4096 * 8

BATCH_SIZES = [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]
# BATCH_SIZES = [128, 256, 512, 1024]

HQQLinearBitBlas.DEFAULT_BATCHSIZE = [1]  # [1] / [8], [16]
# HQQLinearBitBlas.DEFAULT_BATCHSIZE = [16]
HQQLinearBitBlas.check = lambda hqq_layer: True
HQQLinearBitBlas.BIT_TO_DTYPE = {8: "uint8", 4: "uint4", 2: "uint2", 1: "uint1"}

# W_nbits, group_size = 8, in_features
# W_nbits, group_size = 4, 128
# W_nbits, group_size = 2, 128
W_nbits, group_size = 5, 128

W, triton_linear, torchao_linear, bitblas_linear, marlin_linear = gen_data(
    in_features, out_features, W_nbits, group_size
)

print("W_nbits", W_nbits, "group_size", group_size)
for batch_size in BATCH_SIZES:
    x = (
        torch.randn(
            (batch_size, in_features),
            dtype=triton_linear.compute_dtype,
            device="cuda:0",
        )
        / 10.0
    )
    check_valid(x, W, triton_linear)

    ref_time = eval_time(lambda x: torch.matmul(x, W.T), {"x": x.to(W.dtype)})

    if triton_linear is not None:
        triton_time = eval_time(
            lambda x: triton_linear(x), {"x": x.to(triton_linear.compute_dtype)}
        )
        print(
            (batch_size, in_features, out_features),
            "Triton  Speed-up vs. torch.matmul",
            np.round(ref_time / triton_time, 2),
            "time",
            triton_time,
        )

    if torchao_linear is not None:
        torchao_time = eval_time(
            lambda x: torchao_linear(x), {"x": x.to(torchao_linear.compute_dtype)}
        )
        print(
            (batch_size, in_features, out_features),
            "Torchao Speed-up vs. torch.matmul",
            np.round(ref_time / torchao_time, 2),
        )

    if bitblas_linear is not None:
        bitblas_time = eval_time(
            lambda x: bitblas_linear(x), {"x": x.to(bitblas_linear.compute_dtype)}
        )
        print(
            (batch_size, in_features, out_features),
            "Bitblas Speed-up vs. torch.matmul",
            np.round(ref_time / bitblas_time, 2),
        )

    if marlin_linear is not None:
        marlin_time = eval_time(
            lambda x: marlin_linear.forward(x), {"x": x.to(marlin_linear.compute_dtype)}
        )
        print(
            (batch_size, in_features, out_features),
            "Marlin Speed-up vs. torch.matmul",
            np.round(ref_time / marlin_time, 2),
        )

    print("----------------------------------------------")
